@@ -1,32 +1,30 @@
-from google.cloud.bigquery import PartitionRange, RangePartitioning, TimePartitioning
-from google.cloud.bigquery.job import LoadJobConfig, QueryJobConfig
-from google.cloud.exceptions import NotFound
+from typing import TYPE_CHECKING, Optional, Sequence, Union
+
+if TYPE_CHECKING:
+    import pandas as pd
+    import pyspark.sql
+    from google.oauth2.credentials import Credentials
 
-from skt.ye import get_spark
 
 PROJECT_ID = "skt-datahub"
 TEMP_DATASET = "temp_1d"
+TEMP_GCS_BUCKET = "temp-seoul-7d"
 LOCATION = "asia-northeast3"
 CREDENTIALS_SECRET_PATH = "gcp/skt-datahub/dataflow"
 
 
-def get_credentials():
-    from skt.gcp_credentials import get_gcp_credentials
+def get_credentials() -> "Credentials":
+    from .gcp_credentials import get_gcp_credentials
 
     return get_gcp_credentials()
 
 
-def set_default_credentials():
-    import os
-    import uuid
+def get_credentials_json_with_base64() -> str:
+    import base64
 
-    from skt.vault_utils import get_secrets
+    from .gcp_credentials import get_gcp_credentials_json_string
 
-    key_json = get_secrets(CREDENTIALS_SECRET_PATH)["config"]
-    key_json_path = f"/tmp/{str(uuid.uuid4()).replace('-', '_')}"
-    with open(key_json_path, "w") as f:
-        f.write(key_json)
-    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = key_json_path
+    return base64.b64encode(get_gcp_credentials_json_string().encode()).decode()
 
 
 def get_bigquery_storage_client(credentials=None):
@@ -43,13 +41,10 @@ def _get_result_schema(sql, bq_client=None, project_id=PROJECT_ID):
 
     if bq_client is None:
         bq_client = get_bigquery_client(project_id=project_id)
-    job_config = QueryJobConfig(
-        dry_run=True,
-        use_query_cache=False,
-    )
+
+    job_config = QueryJobConfig(dry_run=True, use_query_cache=False)
     query_job = bq_client.query(sql, job_config=job_config)
-    schema = query_job._properties["statistics"]["query"]["schema"]
-    return schema
+    return {"fields": [col.to_api_repr() for col in query_job.schema]}
 
 
 def _get_result_column_type(sql, column, bq_client=None, project_id=PROJECT_ID):
@@ -78,7 +73,12 @@ def _print_query_job_results(query_job):
 
 
 def bq_insert_overwrite_table(sql, destination, project_id=PROJECT_ID):
-    bq = get_bigquery_client(project_id=project_id)
+    from google.cloud.bigquery import QueryJobConfig
+
+    bq = get_bigquery_client(
+        project_id=project_id,
+    )
+
     table = bq.get_table(destination)
     if table.time_partitioning or table.range_partitioning:
         load_query_result_to_partitions(sql, destination, project_id)
@@ -99,7 +99,12 @@ def bq_ctas(sql, destination=None, partition_by=None, clustering_fields=None, pr
     """
     create new table and insert results
     """
-    from google.cloud.bigquery.job import QueryJobConfig
+    from google.cloud.bigquery import (
+        PartitionRange,
+        QueryJobConfig,
+        RangePartitioning,
+        TimePartitioning,
+    )
 
     bq = get_bigquery_client(project_id=project_id)
     if partition_by:
@@ -211,25 +216,11 @@ def _load_bq_ipython_extension(ipython):
     ipython.register_magic_function(_bq_cell_magic, magic_kind="cell", magic_name="bq")
 
 
-def _is_ipython():
-    try:
-        from IPython import get_ipython
-
-        shell = get_ipython().__class__.__name__
-        if shell == "ZMQInteractiveShell":
-            return True  # Jupyter notebook or qtconsole
-        elif shell == "TerminalInteractiveShell":
-            return False  # Terminal running IPython
-        else:
-            return False  # Other type (?)
-    except NameError:
-        return False  # Probably standard Python interpreter
-
-
 def set_gcp_credentials():
     import os
     import tempfile
-    from skt.vault_utils import get_secrets
+
+    from .vault_utils import get_secrets
 
     key = get_secrets(CREDENTIALS_SECRET_PATH)["config"]
     key_file_name = tempfile.mkstemp()[1]
@@ -244,7 +235,9 @@ def import_bigquery_ipython_magic():
 
 
 def load_bigquery_ipython_magic():
-    if _is_ipython():
+    from .notebook_utils import is_in_notebook
+
+    if is_in_notebook():
         from IPython import get_ipython
 
         _load_bq_ipython_extension(get_ipython())
@@ -255,6 +248,7 @@ def load_bigquery_ipython_magic():
 def get_bigquery_client(credentials=None, project_id=PROJECT_ID, location=LOCATION):
     import json
     import os
+
     from google.cloud import bigquery
 
     if credentials is None:
@@ -277,16 +271,20 @@ def get_bigquery_client(credentials=None, project_id=PROJECT_ID, location=LOCATI
         credentials=credentials,
         project=project_id,
         location=location,
-        default_query_job_config=QueryJobConfig(labels=job_config_labels),
+        default_query_job_config=bigquery.QueryJobConfig(
+            labels=job_config_labels,
+        ),
     )
 
 
 def bq_table_exists(table, project_id=PROJECT_ID):
+    from google.cloud import exceptions
+
     try:
         get_bigquery_client(project_id=project_id).get_table(table)
-    except NotFound:
+        return True
+    except exceptions.NotFound:
         return False
-    return True
 
 
 def get_unhidden_partitions(table_name, project_id=PROJECT_ID):
@@ -297,15 +295,21 @@ def get_unhidden_partitions(table_name, project_id=PROJECT_ID):
     return parts
 
 
-def _get_partition_filter(dataset, table_name, partition, project_id=PROJECT_ID):
+def _get_partition_filter(
+    dataset: str,
+    table_name: str,
+    partition: str,
+    project_id: Optional[str] = PROJECT_ID,
+):
     table = get_bigquery_client(project_id=project_id).get_table(f"{dataset}.{table_name}")
-    if "timePartitioning" in table._properties:
-        partition_column_name = table._properties["timePartitioning"]["field"]
+    if table.time_partitioning:
+        partition_column_name = table.time_partitioning.field
         return f"{partition_column_name} = '{partition}'"
-    elif "rangePartitioning" in table._properties:
-        partition_column_name = table._properties["rangePartitioning"]["field"]
+    elif table.range_partitioning:
+        partition_column_name = table.range_partitioning.field
         return f"{partition_column_name} = {partition}"
-    return ""
+    else:
+        return ""
 
 
 def bq_to_pandas(sql, large=False, project_id=PROJECT_ID):
@@ -316,55 +320,120 @@ def bq_to_pandas(sql, large=False, project_id=PROJECT_ID):
     return _bq_table_to_pandas(destination, project_id=project_id)
 
 
-def _bq_table_to_df(
-    dataset, table_name, col_list, partition=None, where=None, spark_session=None, project_id=PROJECT_ID
-):
-    import base64
+def _bq_query_to_df(
+    query: str,
+    spark: Optional["pyspark.sql.SparkSession"] = None,
+    project_id=PROJECT_ID,
+) -> "pyspark.sql.DataFrame":
+    from .ye import get_spark
 
-    from skt.gcp_credentials import get_gcp_credentials_json_string
+    spark = spark or get_spark()
+    spark.conf.set("credentials", get_credentials_json_with_base64())
 
-    if not spark_session:
-        spark_session = get_spark()
-    spark_session.conf.set("spark.sql.execution.arrow.enabled", "false")
-
-    key = get_gcp_credentials_json_string()
     df = (
-        spark_session.read.format("bigquery")
+        spark.read.format("bigquery")
         .option("parentProject", project_id)
         .option("project", project_id)
-        .option("table", f"{project_id}:{dataset}.{table_name}")
-        .option("credentials", base64.b64encode(key.encode()).decode())
+        .option("viewsEnabled", "true")
+        .option("materializationProject", project_id)
+        .option("materializationDataset", TEMP_DATASET)
+        .load(query)
     )
-    if partition:
-        table = get_bigquery_client(project_id=project_id).get_table(f"{dataset}.{table_name}")
-        if "timePartitioning" in table._properties:
-            partition_column_name = table._properties["timePartitioning"]["field"]
-            filter = f"{partition_column_name} = '{partition}'"
-        elif "rangePartitioning" in table._properties:
-            partition_column_name = table._properties["rangePartitioning"]["field"]
-            filter = f"{partition_column_name} = {partition}"
-        else:
-            partition_column_name = None
-        if partition_column_name:
-            df = df.option("filter", filter)
-    df = df.load().select(col_list)
-    if where:
-        df.where(where)
     return df
 
 
+def _bq_table_to_df(
+    dataset: str,
+    table_name: str,
+    col_list: Union[str, Sequence["pyspark.sql.Column"]],
+    partition: Optional[str] = None,
+    where: Optional[Union["pyspark.sql.Column", str]] = None,
+    spark: Optional["pyspark.sql.SparkSession"] = None,
+    project_id: Optional[str] = PROJECT_ID,
+) -> "pyspark.sql.DataFrame":
+    from .ye import get_spark
+
+    if not spark:
+        spark = get_spark()
+
+    spark.conf.set("credentials", get_credentials_json_with_base64())
+
+    df = (
+        spark.read.format("bigquery")
+        .option("parentProject", project_id)
+        .option("project", project_id)
+        .option("table", f"{project_id}:{dataset}.{table_name}")
+    )
+
+    if partition:
+        df_filter = _get_partition_filter(
+            dataset=dataset,
+            table_name=table_name,
+            partition=partition,
+            project_id=project_id,
+        )
+
+        if df_filter:
+            df = df.option("filter", df_filter)
+
+    df = df.load().select(col_list)
+
+    if where:
+        df.where(where)
+
+    return df
+
+
+def bq_to_df(
+    query: str,
+    spark_session: Optional["pyspark.sql.SparkSession"] = None,
+    project_id=PROJECT_ID,
+) -> "pyspark.sql.DataFrame":
+    return _bq_query_to_df(
+        query=query,
+        spark=spark_session,
+        project_id=project_id,
+    )
+
+
 def bq_table_to_df(
-    dataset, table_name, col_list, partition=None, where=None, spark_session=None, project_id=PROJECT_ID
-):
-    return _bq_table_to_df(dataset, table_name, col_list, partition, where, spark_session, project_id=project_id)
+    dataset: str,
+    table_name: str,
+    col_list: Union[str, Sequence["pyspark.sql.Column"]],
+    partition: Optional[str] = None,
+    where: Optional[Union["pyspark.sql.Column", str]] = None,
+    spark_session: Optional["pyspark.sql.SparkSession"] = None,
+    project_id: Optional[str] = PROJECT_ID,
+) -> "pyspark.sql.DataFrame":
+    return _bq_table_to_df(
+        dataset=dataset,
+        table_name=table_name,
+        col_list=col_list,
+        partition=partition,
+        where=where,
+        spark=spark_session,
+        project_id=project_id,
+    )
 
 
 def bq_table_to_parquet(
-    dataset, table_name, output_dir, col_list="*", partition=None, where=None, mode="overwrite", project_id=PROJECT_ID
+    dataset: str,
+    table_name: str,
+    output_dir: str,
+    col_list: Union[str, Sequence[str]] = "*",
+    partition: Optional[str] = None,
+    where: Optional[Union["pyspark.sql.Column", str]] = None,
+    mode: str = "overwrite",
+    project_id: Optional[str] = PROJECT_ID,
 ):
-    bq_table_to_df(dataset, table_name, col_list, partition=partition, where=where, project_id=project_id).write.mode(
-        mode
-    ).parquet(output_dir)
+    bq_table_to_df(
+        dataset=dataset,
+        table_name=table_name,
+        col_list=col_list,
+        partition=partition,
+        where=where,
+        project_id=project_id,
+    ).write.mode(mode).parquet(path=output_dir)
 
 
 def bq_table_to_pandas(dataset, table_name, col_list="*", partition=None, where=None, project_id=PROJECT_ID):
@@ -378,97 +447,139 @@ def bq_table_to_pandas(dataset, table_name, col_list="*", partition=None, where=
         if where:
             conditions.append(f" {where} ")
         query += " AND ".join(conditions)
-    return bq_to_pandas(query=query, project_id=project_id)
+    return bq_to_pandas(sql=query, project_id=project_id)
 
 
 def _df_to_bq_table(
-    df,
-    dataset,
-    table_name,
-    partition=None,
-    partition_field=None,
-    clustering_fields=None,
-    mode="overwrite",
-    project_id=PROJECT_ID,
+    df: "pyspark.sql.DataFrame",
+    dataset: str,
+    table_name: str,
+    partition: Optional[str] = None,
+    partition_field: Optional[str] = None,
+    clustering_fields: Optional[Sequence[str]] = None,
+    mode: str = "overwrite",
+    project_id: Optional[str] = PROJECT_ID,
 ):
-    import base64
-
-    from skt.gcp_credentials import get_gcp_credentials_json_string
-
-    key = get_gcp_credentials_json_string()
     table = f"{dataset}.{table_name}${partition}" if partition else f"{dataset}.{table_name}"
+
+    df.sparkSession.conf.set("credentials", get_credentials_json_with_base64())
+
     df = (
         df.write.format("bigquery")
         .option("parentProject", project_id)
         .option("project", project_id)
-        .option("credentials", base64.b64encode(key.encode()).decode())
         .option("table", table)
-        .option("temporaryGcsBucket", "temp-seoul-7d")
+        .option("temporaryGcsBucket", TEMP_GCS_BUCKET)
     )
     if partition_field:
         df = df.option("partitionField", partition_field)
+
     if clustering_fields:
         df = df.option("clusteredFields", ",".join(clustering_fields))
+
     df.save(mode=mode)
 
 
 def df_to_bq_table(
-    df,
-    dataset,
-    table_name,
-    partition=None,
-    partition_field=None,
-    clustering_fields=None,
-    mode="overwrite",
-    project_id=PROJECT_ID,
+    df: "pyspark.sql.DataFrame",
+    dataset: str,
+    table_name: str,
+    partition: Optional[str] = None,
+    partition_field: Optional[str] = None,
+    clustering_fields: Optional[Sequence[str]] = None,
+    mode: str = "overwrite",
+    project_id: Optional[str] = PROJECT_ID,
 ):
-    _df_to_bq_table(df, dataset, table_name, partition, partition_field, clustering_fields, mode, project_id=project_id)
+    _df_to_bq_table(
+        df=df,
+        dataset=dataset,
+        table_name=table_name,
+        partition=partition,
+        partition_field=partition_field,
+        clustering_fields=clustering_fields,
+        mode=mode,
+        project_id=project_id,
+    )
 
 
 def parquet_to_bq_table(
-    parquet_dir,
-    dataset,
-    table_name,
-    partition=None,
-    partition_field=None,
-    clustering_fields=None,
-    mode="overwrite",
-    project_id=PROJECT_ID,
+    parquet_dir: str,
+    dataset: str,
+    table_name: str,
+    partition: Optional[str] = None,
+    partition_field: Optional[str] = None,
+    clustering_fields: Optional[Sequence[str]] = None,
+    mode: str = "overwrite",
+    project_id: Optional[str] = PROJECT_ID,
+    spark: Optional["pyspark.sql.SparkSession"] = None,
 ):
-    try:
+    from .ye import get_spark
+
+    if not spark:
         spark = get_spark()
-        df = spark.read.format("parquet").load(parquet_dir)
-        _df_to_bq_table(
-            df, dataset, table_name, partition, partition_field, clustering_fields, mode, project_id=project_id
-        )
-    finally:
-        spark.stop()
+
+    df = spark.read.format("parquet").load(parquet_dir)
+    _df_to_bq_table(
+        df=df,
+        dataset=dataset,
+        table_name=table_name,
+        partition=partition,
+        partition_field=partition_field,
+        clustering_fields=clustering_fields,
+        mode=mode,
+        project_id=project_id,
+    )
 
 
 def pandas_to_bq_table(
+    pd_df: "pd.DataFrame",
+    dataset: str,
+    table_name: str,
+    partition: Optional[str] = None,
+    partition_field: Optional[str] = None,
+    clustering_fields: Optional[Sequence[str]] = None,
+    mode: str = "overwrite",
+    project_id: Optional[str] = PROJECT_ID,
+    spark: Optional["pyspark.sql.SparkSession"] = None,
+):
+    from .ye import get_spark
+
+    if not spark:
+        spark = get_spark()
+
+    spark_df = spark.createDataFrame(pd_df)
+    _df_to_bq_table(
+        df=spark_df,
+        dataset=dataset,
+        table_name=table_name,
+        partition=partition,
+        partition_field=partition_field,
+        clustering_fields=clustering_fields,
+        mode=mode,
+        project_id=project_id,
+    )
+
+
+def pandas_to_bq(
     pd_df,
-    dataset,
-    table_name,
+    destination,
     partition=None,
-    partition_field=None,
     clustering_fields=None,
-    mode="overwrite",
+    overwrite=True,
     project_id=PROJECT_ID,
 ):
-    try:
-        spark = get_spark()
-        spark_df = spark.createDataFrame(pd_df)
-        _df_to_bq_table(
-            spark_df, dataset, table_name, partition, partition_field, clustering_fields, mode, project_id=project_id
-        )
-    finally:
-        spark.stop()
+    from google.cloud.bigquery import (
+        PartitionRange,
+        RangePartitioning,
+        TimePartitioning,
+    )
+    from google.cloud.bigquery.job import LoadJobConfig
 
-
-def pandas_to_bq(pd_df, destination, partition=None, clustering_fields=None, overwrite=True, project_id=PROJECT_ID):
     range_partitioning = None
     time_partitioning = None
+
     bq = get_bigquery_client(project_id=project_id)
+
     if bq_table_exists(destination, project_id=project_id):
         target_table = bq.get_table(destination)
         range_partitioning = target_table.range_partitioning
@@ -477,7 +588,7 @@ def pandas_to_bq(pd_df, destination, partition=None, clustering_fields=None, ove
         if partition:
             import datetime
 
-            from pandas.api.types import is_integer_dtype
+            from pandas.core.dtypes.common import is_integer_dtype
 
             if is_integer_dtype(pd_df[partition][0]):
                 range_partitioning = RangePartitioning(
@@ -493,6 +604,8 @@ def pandas_to_bq(pd_df, destination, partition=None, clustering_fields=None, ove
             input_partitions = [(p.strftime("%Y%m%d"), p) for p in set(pd_df[partition])]
         elif range_partitioning:
             input_partitions = [(p, p) for p in set(pd_df[partition])]
+        else:
+            input_partitions = []
         for part, part_val in input_partitions:
             bq.load_table_from_dataframe(
                 dataframe=pd_df[pd_df[partition] == part_val],
@@ -531,22 +644,14 @@ def rdd_to_pandas(func):
     return _rdd_to_pandas
 
 
-def bq_to_df(query, spark_session=None, project_id=PROJECT_ID):
-    temp_table_name = get_temp_table(project_id=project_id)
-    jc = QueryJobConfig(
-        create_disposition="CREATE_IF_NEEDED",
-        write_disposition="WRITE_TRUNCATE",
-        destination=temp_table_name,
-    )
-    bq_client = get_bigquery_client(project_id=project_id)
-    job = bq_client.query(query, job_config=jc)
-    job.result()
-    t = temp_table_name.split(".")
-
-    return _bq_table_to_df(t[1], t[2], "*", spark_session=spark_session, project_id=project_id)
-
-
 def load_query_result_to_table(dest_table, query, part_col_name=None, clustering_fields=None, project_id=PROJECT_ID):
+    from google.cloud.bigquery import (
+        PartitionRange,
+        QueryJobConfig,
+        RangePartitioning,
+        TimePartitioning,
+    )
+
     bq_client = get_bigquery_client(project_id=project_id)
     print(query)
     if bq_table_exists(dest_table, project_id=project_id):
@@ -606,6 +711,28 @@ def get_temp_table(project_id=PROJECT_ID):
     return full_table_id
 
 
+def gcs_to_bq(table_name, bucket_name, bigquery_dataset, bigquery_table, overwrite=True, project_id=PROJECT_ID):
+    from google.cloud.bigquery import LoadJobConfig, SourceFormat, WriteDisposition
+
+    job_config = None
+    if overwrite:
+        job_config = LoadJobConfig(
+            write_disposition=WriteDisposition.WRITE_TRUNCATE,
+            source_format=SourceFormat.PARQUET,
+        )
+    else:
+        job_config = LoadJobConfig(
+            write_disposition=WriteDisposition.WRITE_APPEND,
+            source_format=SourceFormat.PARQUET,
+        )
+
+    bigquery_path = f"{bigquery_dataset}.{bigquery_table}"
+    client = get_bigquery_client(project_id=project_id)
+    uri = f"gs://{bucket_name}/{table_name}/*"
+    load_job = client.load_table_from_uri(uri, bigquery_path, job_config=job_config)
+    load_job.result()
+
+
 def _get_partition_name(table):
     if table.range_partitioning:
         return table.range_partitioning.field
@@ -615,9 +742,30 @@ def _get_partition_name(table):
         raise Exception(f"Table[{table}] is not partitioned.")
 
 
+def get_bigquery_table_partition_info(dataset, table_name, project_id=PROJECT_ID):
+    client = get_bigquery_client(project_id=project_id)
+    bigquery_path = f"{dataset}.{table_name}"
+    table_info = client.get_table(bigquery_path)
+    if table_info.range_partitioning:
+        return {
+            "partitioning": "range_partitioning",
+            "field": table_info.range_partitioning.field,
+            "start": table_info.range_partitioning.range_.start,
+            "end": table_info.range_partitioning.range_.end,
+            "interval": table_info.range_partitioning.range_.interval,
+        }
+    elif table_info.time_partitioning:
+        return {
+            "partitioning": "time_partitioning",
+            "field": table_info.time_partitioning.field,
+            "type": table_info.time_partitioning.type_,
+        }
+    else:
+        return None
+
+
 def load_query_result_to_partitions(query, dest_table, project_id=PROJECT_ID):
-    from google.cloud.bigquery.dataset import DatasetReference
-    from google.cloud.bigquery.table import TableReference
+    from google.cloud.bigquery import DatasetReference, QueryJobConfig, TableReference
 
     bq = get_bigquery_client(project_id=project_id)
     table = bq.get_table(dest_table)
@@ -797,6 +945,8 @@ def storage_object_exist(filename, storage_client=None, bucket_name=None):
 
 
 def storage_object_remove(filename, storage_client=None, bucket_name=None):
+    from google.cloud import exceptions
+
     if storage_client is None:
         raise Exception("storage_client is None")
 
@@ -807,7 +957,7 @@ def storage_object_remove(filename, storage_client=None, bucket_name=None):
     blob = bucket.blob(filename)
     try:
         blob.delete()
-    except NotFound:
+    except exceptions.NotFound:
         pass
 
     if storage_object_exist(filename, storage_client, bucket_name):
